/*
===================================================================================================
Description:
2-D DSMC code for pressure-driven flow of dilute, monoatomic gas through an orifice
Author:       Petr Hotmar, LAPLACE
Work started: May 2013
Based on:     A. Garcia's code (Numerical Methods for Physics) 'dsmcne' - program to simulate planar Couette flow of dilute gas with DSMC
===================================================================================================
*/

#include "Dsmc.h"
#include "SortListSTL.h"
#include "SampList.h"
#include <string>
#include <algorithm>

// in separate files
double rand( long& seed );
double randn( long& seed );
int colider( Matrix& v, Matrix& crmax, double dt, long& seed,
              Matrix& selxtra, double coeff, SortList& sD );
void sorter( Matrix& x, Matrix &v, int & npart, double L, double H, SortList &sD );
void mover( Matrix& x, Matrix& v, int & npart, double L, double H,
            double mpv, double vwall, double dt,
            Matrix& strikes, Matrix& delv, long& seed );
void sampler( Matrix& x, Matrix& v, int & npart, double L, double H,
            SampList& sampD );
// in this file
void initializeAllCellColumns(double pres1, double pres2, int npart, SortList &sD, Matrix& npart_x,  Matrix& pres_x, Matrix& x, Matrix& v, double& eff_num, double& mfp);
void adjustInletOutletCells(SortList &sD, Matrix& npart_x, Matrix& x, Matrix& v, int& npart);
void parseCmdLine(int argc, char *argv[], int& nx, int& ny, int& np, int& nt);
char* getCmdOption(char ** begin, char ** end, const std::string & option);
bool cmdOptionExists(char** begin, char** end, const std::string& option);
void showProgress(int stage, int istep, int nstep);
void linearIndex2MatrixIndices(int jcell, int nx, int ny, int &i, int &j);

namespace Parameters {
  //* Initialize constants  (particle mass, diameter, etc.)
  const double 
    pi = 3.141592654,
    boltz = 1.3806504e-23,      // Boltzmann's constant (J/K)
    torr = 133.322368421053, // Torr (Pa)
    mass = 6.63e-26,         // Mass of argon atom (kg)
    diam = 3.66e-10,         // Effective diameter of argon atom (m)
    temp = 298,              // Room temperature (K)
  //    pres1 = 0.4*torr,        // Pressure in reservoirs (Pa)
    pres1 = 0.008*torr,        // Pressure in reservoirs (Pa)
    pres2 = 2e-3*torr;
  long seed = 1;             // Initial seed for rand (DO NOT USE ZERO)
}

using namespace std;
using namespace Parameters;

int main(int argc, char *argv[]) {

  cout << "\n--- 2-D DSMC code for pressure-driven flow of dilute, monoatomic gas through an orifice. Run 'dsmc -h' for help.\n";

  //* Input parameters
  double
    h = .5e-3, // orifice diameter [m]
    L = 20*h, // domain length
    H = L/2.; // domain heigth
  int 
    ncellx = 100,
    ncelly = 10, // Number of cells in x-dir and y-dir
    npart = 100000, // Initial number of simulation particles (will fluctuate in simulation)
    nstep = 50; // Total number of time steps

  parseCmdLine(argc, argv, ncellx, ncelly, npart, nstep); // over-ride above defaults with user-defined inputs

  double Volume = L*H*H;        // Volume of the system
  double mpv = sqrt(2*kT/mass);  // Thermal speed (most probable initial velocity)
  double vwall = 0; // Wall velocity

  int  ncell=ncellx*ncelly;             // Number of cells
  double dx=L/ncellx, dy=H/ncelly;
  double tau=dx/mpv, dt=0.2*tau;     // Characteristic time (cell advection time) and timestep
  Matrix vrmax(ncell), selxtra(ncell);
  vrmax.set(3*mpv);    // Estimated max rel. speed
  selxtra.set(0.0);    // Used by routine "colider"

  //* Declare object for lists used in sorting
  SortList sortData(L,H,Volume,dx,dy,ncell,ncellx,ncelly,npart);

  //* Initialize object and variables used in statistical sampling
  SampList sampData(ncell,ncellx,ncelly);
  double tsamp = 0;       // Total sampling time

  int colSum = 0;         // Count total collisions
  Matrix strikes(2), strikeSum(2);
  strikeSum.set(0.0);     // Count strikes on each wall

  //* Initialize domain
  Matrix pres_x(ncellx), npart_x(ncellx); // initial pressure and density distribution across domain
  Matrix x(npart,2), v(npart,3); // position and velocity (irrespective of the dimensionality of the domain (2D in this case), the velocities must be in 3D)
  double eff_num, mfp; // weight and mean free path
  initializeAllCellColumns(pres1, pres2, npart, sortData, npart_x, pres_x, x, v, eff_num, mfp);
  double coeff = 0.5*eff_num*pi*diam*diam*dt/(Volume/ncell); // coeff for the acceptance-rejection in the collision routine

  //* PRINT-OUT A SUMMARY
  printf("\nInitial number of particles: %i",npart);
  printf("\nInitial number of particles per cell at inlet, outlet: [npc_in, npc_out]=[%i, %i]",(int) (npart_x(1)/ncelly), (int)(npart_x(ncellx)/ncelly));
  printf("\nCollision grid: [nx*ny=ncell]=[%i*%i=%i]",ncellx,ncelly,ncell);
  printf("\nDomain layout: [L, H, dx, dy]=[%g, %g, %g, %g] orifice diameters",L/h,H/h,dx/h,dy/h);
  printf("\nEach particle represents: %g atoms",eff_num);
  printf("\nLength of collision cell is: %g mean free paths",dx/mfp);
  printf("\nNumber of time steps: %i",nstep);
  printf("\nTime step and simulation time: [dt, t]=[%g, %g] cell advection times",dt/tau, nstep*dt/tau);
  printf("\nReservoir pressures at inlet, outlet: [p1, p2]=[%g, %g] Torr",pres1/torr,pres2/torr);
  printf("\nAccept-reject collision coefficient is: %g",coeff);
  cout << "\n\n";

  if(Debug)
    {
      cout<< "\ndsmc(init)::"; sortData.display();
      x.display("\ndsmc(init)::X=");
      v.display("\ndsmc(init)::V=");
    }

  //***************** TIME LOOP (START) *****************
  for( int istep = 1; istep<=nstep; istep++ ) {

    //* Move all the particles
    showProgress(1,istep,nstep);
    Matrix delv(2);  delv.set(0.0);
    mover( x, v, npart, L, H, mpv, vwall, dt, strikes, delv, seed );
    strikeSum(1) += strikes(1);
    strikeSum(2) += strikes(2);

    //* Sort the particles into cells
    showProgress(2,istep,nstep);
    sorter(x,v,npart,L,H,sortData);

    //* Adjust pressure at intlet/out to conform to BCs
    showProgress(3,istep,nstep);
    adjustInletOutletCells(sortData,npart_x,x,v,npart);

    //* Re-sort
    showProgress(4,istep,nstep);
    sorter(x,v,npart,L,H,sortData);

    //* Evaluate collisions among the particles
    showProgress(5,istep,nstep);
    int col = colider(v,vrmax,dt,seed,selxtra,coeff,sortData);
    colSum += col;  // Increment collision count

    //* After initial transient, accumulate statistical samples
    showProgress(6,istep,nstep);
    if(istep > nstep/10) {
      sampler(x,v,npart,L,H,sampData);
      tsamp += dt;
    }

    //* Periodically display the current progress
    if( (istep%100) < 1 )  {
      cout << "\nDone " << istep << " of " << nstep << " steps; " <<
             colSum << " collisions" << endl;
      cout << "Total wall strikes: " << strikeSum(1) << " (lower) "
           << strikeSum(2) << " (upper)" << endl;
    }
  }
  //***************** TIME LOOP (END) *****************

  //* Normalize the accumulated statistics
  int i, nsamp = sampData.nsamp;
  for( i=1; i<=ncell; i++ ) {
    sampData.ave_n[i] *= (eff_num/(Volume/ncell))/nsamp;
    sampData.ave_ux[i] /= nsamp;
    sampData.ave_uy[i] /= nsamp;
    sampData.ave_uz[i] /= nsamp;
    sampData.ave_T[i] *= mass/(3*boltz*nsamp);
  }

  //* Print out the plotting variables:
  //    xcell, ycell, ave_n, ave_ux, ave_uy, ave_uz, ave_T
  ofstream 
    xcellOut("xcell.txt"), ycellOut("ycell.txt"), 
    ave_uxOut("ave_ux.txt"), ave_uyOut("ave_uy.txt"), ave_uzOut("ave_uz.txt"), 
    ave_nOut("ave_n.txt"), ave_TOut("ave_T.txt");
  int m,n;
  for( i=1; i<=ncell; i++ ) {
    linearIndex2MatrixIndices(i, ncellx, ncelly, m, n); // i -> (m,n)
    xcellOut << (m-0.5)*dx << endl;
    ycellOut << (n-0.5)*dy << endl;
    ave_uxOut << sampData.ave_ux[i] << endl;
    ave_uyOut << sampData.ave_uy[i] << endl;
    ave_uzOut << sampData.ave_uz[i] << endl;
    ave_nOut << sampData.ave_n[i] << endl;
    ave_TOut << sampData.ave_T[i] << endl;
  }

  cout << "\n\nDone.\n";
}


/**************************** SUB-FUNCTIONS ****************************/

void thermalizeParticle(int i, double box[], Matrix & x, Matrix & v)
/***************************************************************************
Assign random position and Maxwellian velocity

  INPUTS:
  - i
  - box = [xmin xmax ymin ymax]

  OUTPUTS:
  - x = position
  - v = velocity

***************************************************************************/
{
  // Assign random positions
  x(i,1) = randRange(box[0],box[1]);
  x(i,2) = randRange(box[2],box[3]);
  // Initial velocities are Maxwell-Boltzmann distributed
  v(i,1) = sqrt(kT/mass) * randn(seed);
  v(i,2) = sqrt(kT/mass) * randn(seed);
  v(i,3) = sqrt(kT/mass) * randn(seed);
}

double randRange(double min, double max)
{
  return min + rand(seed) * (max - min);
}

double linear(double k, double q, double x)
// Initial particle density distribution across the domain is linear
{
  return k*x+q;
}

double round(double d)
{
    return d < 0.0 ? ceil(d - 0.5) : floor(d + 0.5);
}

void initializeAllCellColumns(double pres1, double pres2, int npart, SortList &sD, Matrix& npart_x,  Matrix& pres_x, Matrix& x, Matrix& v, double& eff_num, double& mfp)
/***************************************************************************
Given the axial pressure gradient, fill each axial cell column (ACC) of the domain with the prescribed number of particles.
Assign random positions and Maxwellian velocities.

  INPUTS:
  - p1, p2 = pressure gradient
  - npart = requested particle number in domain
  - sortData = domain geometry

  OUTPUT ARRAYS:
  - npart_x = required particle number in each ACC
  - pres_x = required pressure in each ACC
  - x = position
  - v = velocity

  OUTPUT SCALARS:
  - eff_num = statistical weight
  - mfp = mean free path

***************************************************************************/
{
  //* Initial axial distribution of pressure and particle number in the domain
  double delp=pres2-pres1;
  double Volume=sD.Volume;
  int ncellx=sD.ncellx;

  int i; double xl;
  for (i=1; i<=ncellx; i++) // loop over cell columns
    {
      xl=((double)i-1)/(ncellx-1); // xl=0..1
      pres_x(i)=linear(delp,pres1,xl); // pressure in axial cell column: p(x)=(p2-p1)*xl+p1, xl=0..1
    }
  // As the pressure is a function of position, so are the: density, eff_num, mfp, Kn!
  double 
    ptot=pres_x.sum(), // total pressure (Pa)
    density = ptot/kT,    // number density of argon (m^-3)
    CN=npart/ptot;
  eff_num = density*Volume/npart; // statistical weight
  mfp = Volume/(sqrt(2.0)*pi*diam*diam*npart*eff_num); // mean free path (m)
  
  for (i=1; i<=ncellx; i++)
    npart_x(i)=round(CN*pres_x(i)); // N in axial cell columns
  
  int dN=npart-(int)npart_x.sum(); // check that required number of particles (npart) has been distributed throughout domain
  if (dN)
    {
      printf("\ninitializeAllCellColumns:: dN=%i. Will adjust required number of particles in the left reservoir.",dN);
      npart_x(1)=npart_x(1)+dN;
    }

  //* Assign random positions and velocities to particles
  int k, j=1;
  for (i=1; i<=ncellx; i++)
    {
      xl=(i-1)*sD.dx;
      double column[4]={xl,xl+sD.dx,0,sD.H}; // cell column dimensions
      int N=(int)( npart_x(i) ); // N particles per cell column (convert rounded double to int)
      for( k=1; k<=N; k++,j++ )
	thermalizeParticle(j,column,x,v); // thermalize N particles in cell column
    }

  if(Debug)
    {
      printf("\ninitializeCellColumns::[dN, ptot (Pa), density (m^-3), eff_num, mfp (m)]=[%i, %e, %e, %e, %e]",dN,ptot,density,eff_num,mfp);
      npart_x.display("\ninitializeCellColumns::npart_x="); pres_x.display("\ninitializeCellColumns()::pres_x=");
    }

  // double domain[4]={0,sD.L,0,sD.H};
  // for( i=1; i<=npart; i++ )
  //   thermalizeParticle(i,domain,x,v); // thermalize entire domain
}

void adjustInletOutletCells(SortList &sD, Matrix& npart_x, Matrix& x, Matrix& v, int& npart)
/***************************************************************************
Add/remove particles to/from boundary cells to maintain constant pressure gradient across the domain

  INPUTS:
  - sortData = domain geometry + index arrays
  - npart_x = required particle number in each ACC (corresponding to the given pressure gradient)
  - npart = current particle number in domain

  OUTPUT ARRAYS:
  - x = position
  - v = velocity
  - npart = updated particle number in domain

Note: toRemove array and re-sorting is a simple yet inefficient (memory and time) solution. In the future, re-write sortData arrays
as STL vectors and update them on the fly as particles are removed/added.
***************************************************************************/
{
  int shift, boundary, jcell;
  vector<int> toRemove, random;
  for(shift=0; shift<=1; shift++ ) // Loop over inlet and outlet boundaries
    {
      boundary=1+shift*(sD.ncellx-1); // boundary index (left=1, right=ncellx)
      for(jcell=boundary; jcell<=sD.ncell; jcell=jcell+sD.ncellx ) // Loop over boundary cells
	{
	  int 
	    NP = sD.cell_n[jcell], // no. of particles in cell (present)
	    N0 = (int)(npart_x(boundary)/sD.ncelly),  // no. of particles in cell (required)
	    dN = NP-N0; // dN=N-N0
	  if(Debug) printf("\nadjustInletOutletCells::[jcell,N,N-N0]=[%i,%i,%i]",jcell,NP,dN);
	  
	  if (dN>0) // remove particles
	    {
	      random.clear();
	      for(int i=0; i<NP; ++i) 
		random.push_back(i); // generate a <0,1,2..NP-1> set
	      random_shuffle(random.begin(),random.end()); // do a random shuffle (with no repeates) of the above set
	      for (int p=0; p<abs(dN); p++)
		{
		  int k = random[p]; // pick particle at random out of the current cell ( (int)(rand(seed)*NP) might give repeats )
		  int ip = sD.Xref[ k + sD.index[jcell] ]; // find its index
		  toRemove.push_back(ip); // store index -- will remove particle outside the loop, otherwise cross-references in Xref would no longer be valid
		  if(Debug) {printf("\n  p=%i: [k,ip]=[%i,%i]",p,k,ip); cout.flush();}
		}
	    }
	  else if (dN<0) // add particles
	    {
	      for (int p=0; p<abs(dN); p++)
		{
		  // add row of zeros to x,v; npart++
		  x.addRow(); v.addRow(); // add (append) particle entries
		  npart++; // increase particle counter
		  double cellDims[4]={ sD.cellVertex(jcell,1), sD.cellVertex(jcell,1)+sD.dx, sD.cellVertex(jcell,2), sD.cellVertex(jcell,2)+sD.dy };
		  thermalizeParticle(npart,cellDims,x,v);  // update x,v (x random, v Maxwellian)
		  if(Debug) printf("\n  p=%i: cellDims=[%f,%f,%f,%f]",p,cellDims[0],cellDims[1],cellDims[2],cellDims[3]);
		}
	    }

	} // end for(jcell)
    } // end (for(shift)

  // remove the corresponding row from x,v; npart--
  sort( toRemove.begin(), toRemove.end(), std::greater<int>() ); // sort in descending order
  for( vector<int>::const_iterator i = toRemove.begin(); i != toRemove.end(); ++i) {
    int ip = *i;
    x.removeRow(ip); v.removeRow(ip); // remove particle entries, starting from the highest particle index
    npart--; // decrease particle counter
  }
} // endadjustInletOutletCells()

void parseCmdLine(int argc, char *argv[], int& nx, int& ny, int& np, int& nt)
// parse command line options
{
  // If this section expands, use Boost.Program_options for efficiency
  char* opt;
  char** beg = argv;
  char** end = argv+argc;

  if(cmdOptionExists(beg,end,"-nx"))
      {
	opt = getCmdOption(beg,end, "-nx"); nx=atoi(opt);
      }
  if(cmdOptionExists(beg,end,"-ny"))
      {
	opt = getCmdOption(beg,end, "-ny"); ny=atoi(opt);
      }
  if(cmdOptionExists(beg,end,"-np"))
      {
	opt = getCmdOption(beg,end, "-np"); np=atoi(opt);
      }
  if(cmdOptionExists(beg,end,"-nt"))
      {
	opt = getCmdOption(beg,end, "-nt"); nt=atoi(opt);
      }
  if(cmdOptionExists(beg,end,"-h"))
      {
	cout << "\n--- USE: dsmc [-np <NP>] [-nt <NT>] [-nx <NX>] [-ny <NY>] \nwhere \nnp=number of particles, \nnt=number of time steps, \nnx=number of cells in x-direction, \nny=number of cells in y-direction.\n";
	cout << "\nProgress bar will display MSARCE, where \nM=move, \nS=sort, \nA=adjust particle number, \nR=re-sort, \nC=collide, \nE=evaluate.\n";
	cout << "\nFor debugging set Debug flag to 1 in Dsmc.h.\n";
	exit(0);
      }
}

char* getCmdOption(char ** begin, char ** end, const std::string & option)
{
    char ** itr = std::find(begin, end, option);
    if (itr != end && ++itr != end)
    {
        return *itr;
    }
    return(0);
}

bool cmdOptionExists(char** begin, char** end, const std::string& option)
{
    return std::find(begin, end, option) != end;
}

void showProgress(int stage, int istep, int nstep)
// show which stage of DSMC is currently under way
{
  string s;
  switch(stage)
    {
    case 1: s = "(M     )"; break; // move
    case 2: s = "(MS    )"; break; // sort
    case 3: s = "(MSA   )"; break; // adjust particle number
    case 4: s = "(MSAR  )"; break; // re-sort
    case 5: s = "(MSARC )"; break; // collide
    case 6: s = "(MSARCE)"; break; // evaluate
    default: cout << "\nUNKNOWN stage!\n";
    }
  cerr << "\r--- Step " << istep << " out of " << nstep << " " << s; // use stderr to print immediately (without flushing)
}

void linearIndex2MatrixIndices(int jcell, int nx, int ny, int &i, int &j)
// Convert linear index of a 2D array to standard matrix notation
// IN: jcell, nx, ny
// OUT: i, j
{
  j=(jcell-1)/nx+1;
  i=(jcell-1)%nx+1;
}

int particlePosition2CellAddress(int i, Matrix& x, double L, double H, int ncellx, int ncelly)
// Convert particle position to cell address
{
  int j = (int)(x(i,1)*ncellx/L)+1; j = ( j <= ncellx ) ? j : ncellx;
  int k = (int)(x(i,2)*ncelly/H)+1; k = ( k <= ncelly ) ? k : ncelly;
  return (ncellx*(k-1)+j); // cell address (row-wise ordering)
}
