#include <iostream>
#include <iomanip>
#include <assert.h>  // Defines the assert function.
#include <vector>
#include <algorithm> // for copy
#include <iterator> // for ostream_iterator
#include <string>

using namespace std;

// Use BOOST-UBLAS in the future to supercede all this!
// Using linear addressing (row-wise ordering) to ensure contiguous memory blocks for matrices
class Matrix {

public:

// Default Constructor. Creates a 1 by 1 matrix; sets value to zero. 
  Matrix ():
  nRow_(1), // initialization list
  nCol_(1),
  data_(1)
 {
   set(0.0);                // Set value of data_[0] to 0.0
 }

// Regular Constructor. Creates an nR by nC matrix; sets values to zero.
// If number of columns is not specified, it is set to 1.
 Matrix(int nR, int nC = 1):
 nRow_(nR),
 nCol_(nC),
 data_(nR*nC)
{
  assert(nR > 0 && nC > 0);    // Check that nC and nR both > 0.
  set(0.0);                    // Set values of data_[] to 0.0
}

// Copy Constructor.
// Used when a copy of an object is produced 
// (e.g., passing to a function by value)
Matrix(const Matrix& mat) {
  this->copy(mat);   // Call private copy function.
}

// Destructor. Called when a Matrix object goes out of scope.
~Matrix() {}

// Assignment operator function.
// Overloads the equal sign operator to work with
// Matrix objects.
Matrix& operator=(const Matrix& mat) {
  if( this == &mat ) return *this;  // If two sides equal, do nothing.
  this->copy(mat);                  // Copy right hand side to l.h.s.
  return *this;
}

// Simple "get" functions. Return number of rows or columns.
int nRow() const { return nRow_; }
int nCol() const { return nCol_; }
int size() const { return nRow_*nCol_; }

// Parenthesis operator function.
// Allows access to values of Matrix via (i,j) pair.
// Example: a(1,1) = 2*b(2,3); 
// If column is unspecified, take as 1.
double& operator() (int i, int j = 1) {
  assert(i > 0 && i <= nRow_);          // Bounds checking for rows
  assert(j > 0 && j <= nCol_);          // Bounds checking for columns
  return data_[ nCol_*(i-1) + (j-1) ];  // Access appropriate value
}

// Parenthesis operator function (const version).
const double& operator() (int i, int j = 1) const{
  assert(i > 0 && i <= nRow_);          // Bounds checking for rows
  assert(j > 0 && j <= nCol_);          // Bounds checking for columns
  return data_[ nCol_*(i-1) + (j-1) ];  // Access appropriate value
}

// Set function. Sets all elements of a matrix to a given value.
void set(double value) {
  int i, iData = nRow_*nCol_;
  for( i=0; i<iData; i++ )
    data_[i] = value;
}

// Sum all entries
double sum() {
  int i, iData = nRow_*nCol_;
  double sum=0.;
  for( i=0; i<iData; i++ )
    sum+=data_[i];
  return sum;
} 

// Print path vector to console
 void display(string text = ""){
   int k=1;
   cout << text << "\n[\n";
   for( vector<double>::const_iterator i = data_.begin(); i != data_.end(); ++i, ++k) {
     cout << scientific  << *i << ' ';
     if (k%nCol_==0) cout << endl;
   }
   cout << "]";
 }
 
 void removeRow(int i) {
   if(i<=0 || i>nRow_) {cerr << "\nremoveRow:: Out of bounds (i="<<i<<", expected i=<1.."<<nRow_<<">)! Aborting..."; abort();}
   int row=nCol_*(i-1);
   data_.erase(data_.begin()+row,data_.begin()+row+nCol_);
   nRow_--;
}

 void addRow() {
   // add row of zeros
   data_.insert(data_.end(),nCol_,0);
   nRow_++;
} 
 
//*********************************************************************
private:

// Matrix data.
 int nRow_, nCol_;  // Number of rows, columns
 vector<double> data_;     // Pointer used to allocate memory for data.

// Private copy function.
// Copies values from one Matrix object to another.
void copy(const Matrix& mat) {
  nRow_ = mat.nRow_;
  nCol_ = mat.nCol_;
  int i, iData = nRow_*nCol_;
  for(i = 0; i<iData; i++ )
    data_[i] = mat.data_[i];
}

}; // Class Matrix

