#include "Dsmc.h"

double rand( long& seed );
double randn( long& seed );
int checkBoundaryCrossing(Matrix& x, int i, double L, double H);

void mover( Matrix& x, Matrix& v, int & npart, double L, double H,
		    double mpv, double vwall, double dt,
			Matrix& strikes, Matrix& delv, long& seed ) {

// mover - Function to move particles by free flight
//         Also handles collisions with walls
// Inputs
//    x        Positions of the particles
//    v        Velocities of the particles
//    L        System length
//    mpv      Most probable velocity off the wall
//    vwall    Wall velocities
//    dt       Time step
//    seed     Random number seed
// Outputs
//    x,v      Updated positions and velocities
//    strikes  Number of particles striking each wall
//    delv     Change of y-velocity at each wall     
//    seed     Random number seed

  //* Move all particles pretending walls are absent
  Matrix x0(npart,2);
  x0 = x;            // Remember original position
  int i;
  // Streaming (advection)
  for( i=1; i<= npart; i++ ) {
    x(i,1) = x0(i,1) + v(i,1)*dt;  
    x(i,2) = x0(i,2) + v(i,2)*dt;  
  }

  if(Debug)
    {
      cout << "\nmover (after advection)::npart=" << npart;
      x0.display("\nmover (after advection)::X0="); x.display("\nmover (after advection)::X="); v.display("\nmover (after advection)::V=");
    }

  //* Check each particle to see if it strikes a boundary
  strikes.set(0.0);   delv.set(0.0);
  Matrix ywall(2), xwall(2), direction(2);
  ywall(1) = 0;    ywall(2) = H;   // Positions of walls
  double stdev = mpv/sqrt(2.);
  // Direction of particle leaving wall
  direction(1) = 1;  direction(2) = -1;
  int flag;
  for( i=1; i<=npart; i++ ) { // i and npart are decremented in the loop when particle leaves domain
    //* Test if particle left domain (strikes walls or in/outlet)
    flag = checkBoundaryCrossing(x, i, L, H);
    
    //* If particle strikes a wall, reset its position
    //  and velocity. Record velocity change.
    while( flag == 1 || flag == 2 )	{
      strikes(flag)++;
      //* Reset velocity components as biased Maxwellian,
      //  Exponential dist. in y; Gaussian in x and z
      v(i,1) = stdev*randn(seed);
      v(i,2) = direction(flag)*sqrt(-log(1.-rand(seed))) * mpv;
      v(i,3) = stdev*randn(seed);
      // Time of flight after leaving wall
      double dtr = dt*(x(i,2)-ywall(flag))/(x(i,2)-x0(i,2)); 
      double xwall = x(i,1) - (x(i,1)-x0(i,1))*dtr/dt;
      //* Reset position after leaving wall
      x(i,1) = xwall       + v(i,1)*dtr;
      x(i,2) = ywall(flag) + v(i,2)*dtr;
      //* Re-test if particle left domain (strikes walls or in/outlet)
      flag = checkBoundaryCrossing(x, i, L, H);
      //printf("\nflag=%i:  [dt ywall(flag) x(i,2) x0(i,2)] = [%g %g %g %g]",flag,dt,ywall(flag),x(i,2),x0(i,2));
    }

    //* If particle strikes inlet/outlet, remove particle
    if( flag == 3 || flag == 4 )        {
      // remove the corresponding row from x,x0,v
      x.removeRow(i); // remove particle entries
      x0.removeRow(i);
      v.removeRow(i); 
      npart--; // decrease particle counter
      i--;
    }

  } // end for(ipart)

  if(Debug)
    {
      cout << "\nmover (after boundary treatment)::npart=" << npart;
      x0.display("\nmover (after boundary treatment)::X0="); x.display("\nmover (after boundary treatment)::X="); v.display("\nmover (after boundary treatment)::V=");
    }

}

int checkBoundaryCrossing(Matrix& x, int i, double L, double H)
/***************************************************************************
Check if particle leaves domain, either through walls or through in/outlet

  INPUTS:
  - i = particle index
  - x = position
  - v = velocity

  OUTPUTS:
  - flag = boundary crossing indicator

***************************************************************************/
{
    //* Test if particle strikes walls (in y-direction) or in/outlet (in x-direction)
    int flag = 0;
    if( x(i,2) <= 0 )
      flag=1;       // Particle strikes lower wall
    else if( x(i,2) >= H )
      flag=2;       // Particle strikes upper wall
    if( x(i,1) <= 0 )
      flag=3;       // Particle strikes inlet (and possibly wall)
    else if( x(i,1) >= L )
      flag=4;       // Particle strikes outlet (and possibly wall)

    return flag;
}
